@@ -1,34 +1,34 @@
+use alloc::sync::Arc;
+use alloc::vec::Vec;
+use core::marker::PhantomData;
+use core::ops::{Deref, DerefMut};
+use core::{fmt, mem};
+
+use pki_types::{ServerName, UnixTime};
+
+use super::handy::NoClientSessionStorage;
+use super::hs;
 use crate::builder::ConfigBuilder;
+use crate::client::{EchMode, EchStatus};
 use crate::common_state::{CommonState, Protocol, Side};
-use crate::conn::{ConnectionCommon, ConnectionCore};
+use crate::conn::{ConnectionCore, UnbufferedConnectionCommon};
 use crate::crypto::{CryptoProvider, SupportedKxGroup};
 use crate::enums::{CipherSuite, ProtocolVersion, SignatureScheme};
 use crate::error::Error;
-#[cfg(feature = "logging")]
 use crate::log::trace;
 use crate::msgs::enums::NamedGroup;
 use crate::msgs::handshake::ClientExtension;
 use crate::msgs::persist;
-use crate::sign;
-use crate::suites::{ExtractedSecrets, SupportedCipherSuite};
-use crate::versions;
-use crate::KeyLog;
-#[cfg(feature = "ring")]
+use crate::suites::SupportedCipherSuite;
+#[cfg(feature = "std")]
+use crate::time_provider::DefaultTimeProvider;
+use crate::time_provider::TimeProvider;
+use crate::unbuffered::{EncryptError, TransmitTlsData};
+#[cfg(feature = "std")]
 use crate::WantsVerifier;
-use crate::{verify, WantsVersions};
-
-use super::handy::{ClientSessionMemoryCache, NoClientSessionStorage};
-use super::hs;
-
-use pki_types::ServerName;
-
-use alloc::sync::Arc;
-use alloc::vec::Vec;
-use core::fmt;
-use core::marker::PhantomData;
-use core::mem;
-use core::ops::{Deref, DerefMut};
-use std::io;
+use crate::{compress, sign, verify, versions, KeyLog, WantsVersions};
+#[cfg(doc)]
+use crate::{crypto, DistinguishedName};
 
 /// A trait for the ability to store client session data, so that sessions
 /// can be resumed in future connections.
@@ -103,7 +103,7 @@ pub trait ResolvesClientCert: fmt::Debug + Send + Sync {
     /// decide on a client certificate the server is likely to accept. If
     /// the list is empty, the client should send whatever certificate it
     /// has. The hints are expected to be DER-encoded X.500 distinguished names,
-    /// per [RFC 5280 A.1]. See [`crate::DistinguishedName`] for more information
+    /// per [RFC 5280 A.1]. See [`DistinguishedName`] for more information
     /// on decoding with external crates like `x509-parser`.
     ///
     /// `sigschemes` is the list of the [`SignatureScheme`]s the server
@@ -134,6 +134,12 @@ pub trait ResolvesClientCert: fmt::Debug + Send + Sync {
 /// These must be created via the [`ClientConfig::builder()`] or [`ClientConfig::builder_with_provider()`]
 /// function.
 ///
+/// Note that using [`ConfigBuilder<ClientConfig, WantsVersions>::with_ech()`] will produce a common
+/// configuration specific to the provided [`crate::client::EchConfig`] that may not be appropriate
+/// for all connections made by the program. In this case the configuration should only be shared
+/// by connections intended for domains that offer the provided [`crate::client::EchConfig`] in
+/// their DNS zone.
+///
 /// # Defaults
 ///
 /// * [`ClientConfig::max_fragment_size`]: the default is `None` (meaning 16kB).
@@ -141,13 +147,13 @@ pub trait ResolvesClientCert: fmt::Debug + Send + Sync {
 ///    ids or tickets, with a max of eight tickets per server.
 /// * [`ClientConfig::alpn_protocols`]: the default is empty -- no ALPN protocol is negotiated.
 /// * [`ClientConfig::key_log`]: key material is not logged.
+/// * [`ClientConfig::cert_decompressors`]: depends on the crate features, see [`compress::default_cert_decompressors()`].
+/// * [`ClientConfig::cert_compressors`]: depends on the crate features, see [`compress::default_cert_compressors()`].
+/// * [`ClientConfig::cert_compression_cache`]: caches the most recently used 4 compressions
 ///
 /// [`RootCertStore`]: crate::RootCertStore
-#[derive(Debug)]
+#[derive(Clone, Debug)]
 pub struct ClientConfig {
-    /// Source of randomness and other crypto.
-    pub(super) provider: Arc<CryptoProvider>,
-
     /// Which ALPN protocols we include in our client hello.
     /// If empty, no ALPN extension is sent.
     pub alpn_protocols: Vec<Vec<u8>>,
@@ -171,19 +177,12 @@ pub struct ClientConfig {
     /// How to decide what client auth certificate/keys to use.
     pub client_auth_cert_resolver: Arc<dyn ResolvesClientCert>,
 
-    /// Supported versions, in no particular order.  The default
-    /// is all supported versions.
-    pub(super) versions: versions::EnabledVersions,
-
     /// Whether to send the Server Name Indication (SNI) extension
     /// during the client handshake.
     ///
     /// The default is true.
     pub enable_sni: bool,
 
-    /// How to verify the server certificate chain.
-    pub(super) verifier: Arc<dyn verify::ServerCertVerifier>,
-
     /// How to output key material for debugging.  The default
     /// does nothing.
     pub key_log: Arc<dyn KeyLog>,
@@ -197,71 +196,104 @@ pub struct ClientConfig {
     ///
     /// The default is false.
     pub enable_early_data: bool,
-}
 
-/// What mechanisms to support for resuming a TLS 1.2 session.
-#[derive(Clone, Copy, Debug, PartialEq)]
-pub enum Tls12Resumption {
-    /// Disable 1.2 resumption.
-    Disabled,
-    /// Support 1.2 resumption using session ids only.
-    SessionIdOnly,
-    /// Support 1.2 resumption using session ids or RFC 5077 tickets.
+    /// If set to `true`, requires the server to support the extended
+    /// master secret extraction method defined in [RFC 7627].
     ///
-    /// See[^1] for why you might like to disable RFC 5077 by instead choosing the `SessionIdOnly`
-    /// option. Note that TLS 1.3 tickets do not have those issues.
+    /// The default is `true` if the `fips` crate feature is enabled,
+    /// `false` otherwise.
     ///
-    /// [^1]: <https://words.filippo.io/we-need-to-talk-about-session-tickets/>
-    SessionIdOrTickets,
-}
+    /// It must be set to `true` to meet FIPS requirement mentioned in section
+    /// **D.Q Transition of the TLS 1.2 KDF to Support the Extended Master
+    /// Secret** from [FIPS 140-3 IG.pdf].
+    ///
+    /// [RFC 7627]: https://datatracker.ietf.org/doc/html/rfc7627
+    /// [FIPS 140-3 IG.pdf]: https://csrc.nist.gov/csrc/media/Projects/cryptographic-module-validation-program/documents/fips%20140-3/FIPS%20140-3%20IG.pdf
+    #[cfg(feature = "tls12")]
+    pub require_ems: bool,
 
-impl Clone for ClientConfig {
-    fn clone(&self) -> Self {
-        Self {
-            provider: Arc::<CryptoProvider>::clone(&self.provider),
-            resumption: self.resumption.clone(),
-            alpn_protocols: self.alpn_protocols.clone(),
-            max_fragment_size: self.max_fragment_size,
-            client_auth_cert_resolver: Arc::clone(&self.client_auth_cert_resolver),
-            versions: self.versions,
-            enable_sni: self.enable_sni,
-            verifier: Arc::clone(&self.verifier),
-            key_log: Arc::clone(&self.key_log),
-            enable_secret_extraction: self.enable_secret_extraction,
-            enable_early_data: self.enable_early_data,
-        }
-    }
+    /// Provides the current system time
+    pub time_provider: Arc<dyn TimeProvider>,
+
+    /// Source of randomness and other crypto.
+    pub(super) provider: Arc<CryptoProvider>,
+
+    /// Supported versions, in no particular order.  The default
+    /// is all supported versions.
+    pub(super) versions: versions::EnabledVersions,
+
+    /// How to verify the server certificate chain.
+    pub(super) verifier: Arc<dyn verify::ServerCertVerifier>,
+
+    /// How to decompress the server's certificate chain.
+    ///
+    /// If this is non-empty, the [RFC8779] certificate compression
+    /// extension is offered, and any compressed certificates are
+    /// transparently decompressed during the handshake.
+    ///
+    /// This only applies to TLS1.3 connections.  It is ignored for
+    /// TLS1.2 connections.
+    ///
+    /// [RFC8779]: https://datatracker.ietf.org/doc/rfc8879/
+    pub cert_decompressors: Vec<&'static dyn compress::CertDecompressor>,
+
+    /// How to compress the client's certificate chain.
+    ///
+    /// If a server supports this extension, and advertises support
+    /// for one of the compression algorithms included here, the
+    /// client certificate will be compressed according to [RFC8779].
+    ///
+    /// This only applies to TLS1.3 connections.  It is ignored for
+    /// TLS1.2 connections.
+    ///
+    /// [RFC8779]: https://datatracker.ietf.org/doc/rfc8879/
+    pub cert_compressors: Vec<&'static dyn compress::CertCompressor>,
+
+    /// Caching for compressed certificates.
+    ///
+    /// This is optional: [`compress::CompressionCache::Disabled`] gives
+    /// a cache that does no caching.
+    pub cert_compression_cache: Arc<compress::CompressionCache>,
+
+    /// How to offer Encrypted Client Hello (ECH). The default is to not offer ECH.
+    pub(super) ech_mode: Option<EchMode>,
 }
 
 impl ClientConfig {
-    /// Create a builder for a client configuration with the default
-    /// [`CryptoProvider`]: [`crate::crypto::ring::default_provider`] and safe ciphersuite and
-    /// protocol defaults.
+    /// Create a builder for a client configuration with
+    /// [the process-default `CryptoProvider`][CryptoProvider#using-the-per-process-default-cryptoprovider]
+    /// and safe protocol version defaults.
     ///
     /// For more information, see the [`ConfigBuilder`] documentation.
-    #[cfg(feature = "ring")]
+    #[cfg(feature = "std")]
     pub fn builder() -> ConfigBuilder<Self, WantsVerifier> {
-        // Safety: we know the *ring* provider's ciphersuites are compatible with the safe default protocol versions.
-        Self::builder_with_provider(crate::crypto::ring::default_provider().into())
-            .with_safe_default_protocol_versions()
-            .unwrap()
+        Self::builder_with_protocol_versions(versions::DEFAULT_VERSIONS)
     }
 
-    /// Create a builder for a client configuration with the default
-    /// [`CryptoProvider`]: [`crate::crypto::ring::default_provider`], safe ciphersuite defaults and
-    /// the provided protocol versions.
+    /// Create a builder for a client configuration with
+    /// [the process-default `CryptoProvider`][CryptoProvider#using-the-per-process-default-cryptoprovider]
+    /// and the provided protocol versions.
     ///
-    /// Panics if provided an empty slice of supported versions.
+    /// Panics if
+    /// - the supported versions are not compatible with the provider (eg.
+    ///   the combination of ciphersuites supported by the provider and supported
+    ///   versions lead to zero cipher suites being usable),
+    /// - if a `CryptoProvider` cannot be resolved using a combination of
+    ///   the crate features and process default.
     ///
     /// For more information, see the [`ConfigBuilder`] documentation.
-    #[cfg(feature = "ring")]
+    #[cfg(feature = "std")]
     pub fn builder_with_protocol_versions(
         versions: &[&'static versions::SupportedProtocolVersion],
     ) -> ConfigBuilder<Self, WantsVerifier> {
-        // Safety: we know the *ring* provider's ciphersuites are compatible with all protocol version choices.
-        Self::builder_with_provider(crate::crypto::ring::default_provider().into())
-            .with_protocol_versions(versions)
-            .unwrap()
+        // Safety assumptions:
+        // 1. that the provider has been installed (explicitly or implicitly)
+        // 2. that the process-level default provider is usable with the supplied protocol versions.
+        Self::builder_with_provider(Arc::clone(
+            CryptoProvider::get_default_or_install_from_crate_features(),
+        ))
+        .with_protocol_versions(versions)
+        .unwrap()
     }
 
     /// Create a builder for a client configuration with a specific [`CryptoProvider`].
@@ -272,14 +304,76 @@ impl ClientConfig {
     /// version is not supported by the provider's ciphersuites.
     ///
     /// For more information, see the [`ConfigBuilder`] documentation.
+    #[cfg(feature = "std")]
     pub fn builder_with_provider(
         provider: Arc<CryptoProvider>,
     ) -> ConfigBuilder<Self, WantsVersions> {
         ConfigBuilder {
-            state: WantsVersions { provider },
+            state: WantsVersions {
+                provider,
+                time_provider: Arc::new(DefaultTimeProvider),
+            },
             side: PhantomData,
         }
     }
+    /// Create a builder for a client configuration with no default implementation details.
+    ///
+    /// This API must be used by `no_std` users.
+    ///
+    /// You must provide a specific [`TimeProvider`].
+    ///
+    /// You must provide a specific [`CryptoProvider`].
+    ///
+    /// This will use the provider's configured ciphersuites. You must additionally choose
+    /// which protocol versions to enable, using `with_protocol_versions` or
+    /// `with_safe_default_protocol_versions` and handling the `Result` in case a protocol
+    /// version is not supported by the provider's ciphersuites.
+    ///
+    /// For more information, see the [`ConfigBuilder`] documentation.
+    pub fn builder_with_details(
+        provider: Arc<CryptoProvider>,
+        time_provider: Arc<dyn TimeProvider>,
+    ) -> ConfigBuilder<Self, WantsVersions> {
+        ConfigBuilder {
+            state: WantsVersions {
+                provider,
+                time_provider,
+            },
+            side: PhantomData,
+        }
+    }
+
+    /// Return true if connections made with this `ClientConfig` will
+    /// operate in FIPS mode.
+    ///
+    /// This is different from [`CryptoProvider::fips()`]: [`CryptoProvider::fips()`]
+    /// is concerned only with cryptography, whereas this _also_ covers TLS-level
+    /// configuration that NIST recommends, as well as ECH HPKE suites if applicable.
+    pub fn fips(&self) -> bool {
+        let mut is_fips = self.provider.fips();
+
+        #[cfg(feature = "tls12")]
+        {
+            is_fips = is_fips && self.require_ems
+        }
+
+        if let Some(ech_mode) = &self.ech_mode {
+            is_fips = is_fips && ech_mode.fips();
+        }
+
+        is_fips
+    }
+
+    /// Return the crypto provider used to construct this client configuration.
+    pub fn crypto_provider(&self) -> &Arc<CryptoProvider> {
+        &self.provider
+    }
+
+    /// Access configuration options whose use is dangerous and requires
+    /// extra care.
+    pub fn dangerous(&mut self) -> danger::DangerousClientConfig<'_> {
+        danger::DangerousClientConfig { cfg: self }
+    }
 
     /// We support a given TLS version if it's quoted in the configured
     /// versions *and* at least one ciphersuite for this version is
@@ -293,6 +387,7 @@ impl ClientConfig {
                 .any(|cs| cs.version().version == v)
     }
 
+    #[cfg(feature = "std")]
     pub(crate) fn supports_protocol(&self, proto: Protocol) -> bool {
         self.provider
             .cipher_suites
@@ -300,12 +395,6 @@ impl ClientConfig {
             .any(|cs| cs.usable_for_protocol(proto))
     }
 
-    /// Access configuration options whose use is dangerous and requires
-    /// extra care.
-    pub fn dangerous(&mut self) -> danger::DangerousClientConfig<'_> {
-        danger::DangerousClientConfig { cfg: self }
-    }
-
     pub(super) fn find_cipher_suite(&self, suite: CipherSuite) -> Option<SupportedCipherSuite> {
         self.provider
             .cipher_suites
@@ -314,12 +403,22 @@ impl ClientConfig {
             .find(|&scs| scs.suite() == suite)
     }
 
-    pub(super) fn find_kx_group(&self, group: NamedGroup) -> Option<&'static dyn SupportedKxGroup> {
+    pub(super) fn find_kx_group(
+        &self,
+        group: NamedGroup,
+        version: ProtocolVersion,
+    ) -> Option<&'static dyn SupportedKxGroup> {
         self.provider
             .kx_groups
             .iter()
             .copied()
-            .find(|skxg| skxg.name() == group)
+            .find(|skxg| skxg.usable_for_version(version) && skxg.name() == group)
+    }
+
+    pub(super) fn current_time(&self) -> Result<UnixTime, Error> {
+        self.time_provider
+            .current_time()
+            .ok_or(Error::FailedToGetCurrentTime)
     }
 }
 
@@ -327,7 +426,7 @@ impl ClientConfig {
 #[derive(Clone, Debug)]
 pub struct Resumption {
     /// How we store session data or tickets. The default is to use an in-memory
-    /// [ClientSessionMemoryCache].
+    /// [super::handy::ClientSessionMemoryCache].
     pub(super) store: Arc<dyn ClientSessionStore>,
 
     /// What mechanism is used for resuming a TLS 1.2 session.
@@ -339,9 +438,10 @@ impl Resumption {
     ///
     /// This is the default `Resumption` choice, and enables resuming a TLS 1.2 session with
     /// a session id or RFC 5077 ticket.
+    #[cfg(feature = "std")]
     pub fn in_memory_sessions(num: usize) -> Self {
         Self {
-            store: Arc::new(ClientSessionMemoryCache::new(num)),
+            store: Arc::new(super::handy::ClientSessionMemoryCache::new(num)),
             tls12_resumption: Tls12Resumption::SessionIdOrTickets,
         }
     }
@@ -366,7 +466,8 @@ impl Resumption {
 
     /// Configure whether TLS 1.2 sessions may be resumed, and by what mechanism.
     ///
-    /// This is meaningless if you've disabled resumption entirely.
+    /// This is meaningless if you've disabled resumption entirely, which is the case in `no-std`
+    /// contexts.
     pub fn tls12_resumption(mut self, tls12: Tls12Resumption) -> Self {
         self.tls12_resumption = tls12;
         self
@@ -377,10 +478,32 @@ impl Default for Resumption {
     /// Create an in-memory session store resumption with up to 256 server names, allowing
     /// a TLS 1.2 session to resume with a session id or RFC 5077 ticket.
     fn default() -> Self {
-        Self::in_memory_sessions(256)
+        #[cfg(feature = "std")]
+        let ret = Self::in_memory_sessions(256);
+
+        #[cfg(not(feature = "std"))]
+        let ret = Self::disabled();
+
+        ret
     }
 }
 
+/// What mechanisms to support for resuming a TLS 1.2 session.
+#[derive(Clone, Copy, Debug, PartialEq)]
+pub enum Tls12Resumption {
+    /// Disable 1.2 resumption.
+    Disabled,
+    /// Support 1.2 resumption using session ids only.
+    SessionIdOnly,
+    /// Support 1.2 resumption using session ids or RFC 5077 tickets.
+    ///
+    /// See[^1] for why you might like to disable RFC 5077 by instead choosing the `SessionIdOnly`
+    /// option. Note that TLS 1.3 tickets do not have those issues.
+    ///
+    /// [^1]: <https://words.filippo.io/we-need-to-talk-about-session-tickets/>
+    SessionIdOrTickets,
+}
+
 /// Container for unsafe APIs
 pub(super) mod danger {
     use alloc::sync::Arc;
@@ -395,7 +518,7 @@ pub(super) mod danger {
         pub cfg: &'a mut ClientConfig,
     }
 
-    impl<'a> DangerousClientConfig<'a> {
+    impl DangerousClientConfig<'_> {
         /// Overrides the default `ServerCertVerifier` with something else.
         pub fn set_certificate_verifier(&mut self, verifier: Arc<dyn ServerCertVerifier>) {
             self.cfg.verifier = verifier;
@@ -430,6 +553,7 @@ impl EarlyData {
         matches!(self.state, EarlyDataState::Ready | EarlyDataState::Accepted)
     }
 
+    #[cfg(feature = "std")]
     fn is_accepted(&self) -> bool {
         matches!(
             self.state,
@@ -462,7 +586,7 @@ impl EarlyData {
         }
     }
 
-    fn check_write(&mut self, sz: usize) -> io::Result<usize> {
+    fn check_write_opt(&mut self, sz: usize) -> Option<usize> {
         match self.state {
             EarlyDataState::Disabled => unreachable!(),
             EarlyDataState::Ready | EarlyDataState::Accepted => {
@@ -473,165 +597,205 @@ impl EarlyData {
                     sz
                 };
 
-                Ok(take)
-            }
-            EarlyDataState::Rejected | EarlyDataState::AcceptedFinished => {
-                Err(io::Error::from(io::ErrorKind::InvalidInput))
+                Some(take)
             }
+            EarlyDataState::Rejected | EarlyDataState::AcceptedFinished => None,
+        }
+    }
+}
+
+#[cfg(feature = "std")]
+mod connection {
+    use alloc::sync::Arc;
+    use alloc::vec::Vec;
+    use core::fmt;
+    use core::ops::{Deref, DerefMut};
+    use std::io;
+
+    use pki_types::ServerName;
+
+    use super::ClientConnectionData;
+    use crate::client::EchStatus;
+    use crate::common_state::Protocol;
+    use crate::conn::{ConnectionCommon, ConnectionCore};
+    use crate::error::Error;
+    use crate::suites::ExtractedSecrets;
+    use crate::ClientConfig;
+
+    /// Stub that implements io::Write and dispatches to `write_early_data`.
+    pub struct WriteEarlyData<'a> {
+        sess: &'a mut ClientConnection,
+    }
+
+    impl<'a> WriteEarlyData<'a> {
+        fn new(sess: &'a mut ClientConnection) -> Self {
+            WriteEarlyData { sess }
+        }
+
+        /// How many bytes you may send.  Writes will become short
+        /// once this reaches zero.
+        pub fn bytes_left(&self) -> usize {
+            self.sess
+                .inner
+                .core
+                .data
+                .early_data
+                .bytes_left()
         }
     }
 
-    fn bytes_left(&self) -> usize {
-        self.left
-    }
-}
+    impl io::Write for WriteEarlyData<'_> {
+        fn write(&mut self, buf: &[u8]) -> io::Result<usize> {
+            self.sess.write_early_data(buf)
+        }
 
-/// Stub that implements io::Write and dispatches to `write_early_data`.
-pub struct WriteEarlyData<'a> {
-    sess: &'a mut ClientConnection,
-}
-
-impl<'a> WriteEarlyData<'a> {
-    fn new(sess: &'a mut ClientConnection) -> WriteEarlyData<'a> {
-        WriteEarlyData { sess }
-    }
-
-    /// How many bytes you may send.  Writes will become short
-    /// once this reaches zero.
-    pub fn bytes_left(&self) -> usize {
-        self.sess
-            .inner
-            .core
-            .data
-            .early_data
-            .bytes_left()
-    }
-}
-
-impl<'a> io::Write for WriteEarlyData<'a> {
-    fn write(&mut self, buf: &[u8]) -> io::Result<usize> {
-        self.sess.write_early_data(buf)
-    }
-
-    fn flush(&mut self) -> io::Result<()> {
-        Ok(())
-    }
-}
-
-/// This represents a single TLS client connection.
-pub struct ClientConnection {
-    inner: ConnectionCommon<ClientConnectionData>,
-}
-
-impl fmt::Debug for ClientConnection {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        f.debug_struct("ClientConnection")
-            .finish()
-    }
-}
-
-impl ClientConnection {
-    /// Make a new ClientConnection.  `config` controls how
-    /// we behave in the TLS protocol, `name` is the
-    /// name of the server we want to talk to.
-    pub fn new(config: Arc<ClientConfig>, name: ServerName<'static>) -> Result<Self, Error> {
-        Ok(Self {
-            inner: ConnectionCore::for_client(config, name, Vec::new(), Protocol::Tcp)?.into(),
-        })
-    }
-
-    /// Returns an `io::Write` implementer you can write bytes to
-    /// to send TLS1.3 early data (a.k.a. "0-RTT data") to the server.
-    ///
-    /// This returns None in many circumstances when the capability to
-    /// send early data is not available, including but not limited to:
-    ///
-    /// - The server hasn't been talked to previously.
-    /// - The server does not support resumption.
-    /// - The server does not support early data.
-    /// - The resumption data for the server has expired.
-    ///
-    /// The server specifies a maximum amount of early data.  You can
-    /// learn this limit through the returned object, and writes through
-    /// it will process only this many bytes.
-    ///
-    /// The server can choose not to accept any sent early data --
-    /// in this case the data is lost but the connection continues.  You
-    /// can tell this happened using `is_early_data_accepted`.
-    pub fn early_data(&mut self) -> Option<WriteEarlyData> {
-        if self
-            .inner
-            .core
-            .data
-            .early_data
-            .is_enabled()
-        {
-            Some(WriteEarlyData::new(self))
-        } else {
-            None
+        fn flush(&mut self) -> io::Result<()> {
+            Ok(())
         }
     }
 
-    /// Returns True if the server signalled it will process early data.
-    ///
-    /// If you sent early data and this returns false at the end of the
-    /// handshake then the server will not process the data.  This
-    /// is not an error, but you may wish to resend the data.
-    pub fn is_early_data_accepted(&self) -> bool {
-        self.inner.core.is_early_data_accepted()
+    impl super::EarlyData {
+        fn check_write(&mut self, sz: usize) -> io::Result<usize> {
+            self.check_write_opt(sz)
+                .ok_or_else(|| io::Error::from(io::ErrorKind::InvalidInput))
+        }
+
+        fn bytes_left(&self) -> usize {
+            self.left
+        }
     }
 
-    /// Extract secrets, so they can be used when configuring kTLS, for example.
-    /// Should be used with care as it exposes secret key material.
-    pub fn dangerous_extract_secrets(self) -> Result<ExtractedSecrets, Error> {
-        self.inner.dangerous_extract_secrets()
+    /// This represents a single TLS client connection.
+    pub struct ClientConnection {
+        inner: ConnectionCommon<ClientConnectionData>,
     }
 
-    fn write_early_data(&mut self, data: &[u8]) -> io::Result<usize> {
-        self.inner
-            .core
-            .data
-            .early_data
-            .check_write(data.len())
-            .map(|sz| {
-                self.inner
-                    .send_early_plaintext(&data[..sz])
+    impl fmt::Debug for ClientConnection {
+        fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+            f.debug_struct("ClientConnection")
+                .finish()
+        }
+    }
+
+    impl ClientConnection {
+        /// Make a new ClientConnection.  `config` controls how
+        /// we behave in the TLS protocol, `name` is the
+        /// name of the server we want to talk to.
+        pub fn new(config: Arc<ClientConfig>, name: ServerName<'static>) -> Result<Self, Error> {
+            Ok(Self {
+                inner: ConnectionCore::for_client(config, name, Vec::new(), Protocol::Tcp)?.into(),
             })
+        }
+
+        /// Returns an `io::Write` implementer you can write bytes to
+        /// to send TLS1.3 early data (a.k.a. "0-RTT data") to the server.
+        ///
+        /// This returns None in many circumstances when the capability to
+        /// send early data is not available, including but not limited to:
+        ///
+        /// - The server hasn't been talked to previously.
+        /// - The server does not support resumption.
+        /// - The server does not support early data.
+        /// - The resumption data for the server has expired.
+        ///
+        /// The server specifies a maximum amount of early data.  You can
+        /// learn this limit through the returned object, and writes through
+        /// it will process only this many bytes.
+        ///
+        /// The server can choose not to accept any sent early data --
+        /// in this case the data is lost but the connection continues.  You
+        /// can tell this happened using `is_early_data_accepted`.
+        pub fn early_data(&mut self) -> Option<WriteEarlyData<'_>> {
+            if self
+                .inner
+                .core
+                .data
+                .early_data
+                .is_enabled()
+            {
+                Some(WriteEarlyData::new(self))
+            } else {
+                None
+            }
+        }
+
+        /// Returns True if the server signalled it will process early data.
+        ///
+        /// If you sent early data and this returns false at the end of the
+        /// handshake then the server will not process the data.  This
+        /// is not an error, but you may wish to resend the data.
+        pub fn is_early_data_accepted(&self) -> bool {
+            self.inner.core.is_early_data_accepted()
+        }
+
+        /// Extract secrets, so they can be used when configuring kTLS, for example.
+        /// Should be used with care as it exposes secret key material.
+        pub fn dangerous_extract_secrets(self) -> Result<ExtractedSecrets, Error> {
+            self.inner.dangerous_extract_secrets()
+        }
+
+        /// Return the connection's Encrypted Client Hello (ECH) status.
+        pub fn ech_status(&self) -> EchStatus {
+            self.inner.core.data.ech_status
+        }
+
+        /// Return true if the connection was made with a `ClientConfig` that is FIPS compatible.
+        ///
+        /// This is different from [`crate::crypto::CryptoProvider::fips()`]:
+        /// it is concerned only with cryptography, whereas this _also_ covers TLS-level
+        /// configuration that NIST recommends, as well as ECH HPKE suites if applicable.
+        pub fn fips(&self) -> bool {
+            self.inner.core.data.fips
+        }
+
+        fn write_early_data(&mut self, data: &[u8]) -> io::Result<usize> {
+            self.inner
+                .core
+                .data
+                .early_data
+                .check_write(data.len())
+                .map(|sz| {
+                    self.inner
+                        .send_early_plaintext(&data[..sz])
+                })
+        }
     }
-}
 
-impl Deref for ClientConnection {
-    type Target = ConnectionCommon<ClientConnectionData>;
+    impl Deref for ClientConnection {
+        type Target = ConnectionCommon<ClientConnectionData>;
 
-    fn deref(&self) -> &Self::Target {
-        &self.inner
+        fn deref(&self) -> &Self::Target {
+            &self.inner
+        }
     }
-}
 
-impl DerefMut for ClientConnection {
-    fn deref_mut(&mut self) -> &mut Self::Target {
-        &mut self.inner
+    impl DerefMut for ClientConnection {
+        fn deref_mut(&mut self) -> &mut Self::Target {
+            &mut self.inner
+        }
     }
-}
 
-#[doc(hidden)]
-impl<'a> TryFrom<&'a mut crate::Connection> for &'a mut ClientConnection {
-    type Error = ();
+    #[doc(hidden)]
+    impl<'a> TryFrom<&'a mut crate::Connection> for &'a mut ClientConnection {
+        type Error = ();
 
-    fn try_from(value: &'a mut crate::Connection) -> Result<Self, Self::Error> {
-        use crate::Connection::*;
-        match value {
-            Client(conn) => Ok(conn),
-            Server(_) => Err(()),
+        fn try_from(value: &'a mut crate::Connection) -> Result<Self, Self::Error> {
+            use crate::Connection::*;
+            match value {
+                Client(conn) => Ok(conn),
+                Server(_) => Err(()),
+            }
+        }
+    }
+
+    impl From<ClientConnection> for crate::Connection {
+        fn from(conn: ClientConnection) -> Self {
+            Self::Client(conn)
         }
     }
 }
-
-impl From<ClientConnection> for crate::Connection {
-    fn from(conn: ClientConnection) -> Self {
-        Self::Client(conn)
-    }
-}
+#[cfg(feature = "std")]
+pub use connection::{ClientConnection, WriteEarlyData};
 
 impl ConnectionCore<ClientConnectionData> {
     pub(crate) fn for_client(
@@ -645,26 +809,144 @@ impl ConnectionCore<ClientConnectionData> {
         common_state.protocol = proto;
         common_state.enable_secret_extraction = config.enable_secret_extraction;
         let mut data = ClientConnectionData::new();
+        data.fips = config.fips();
 
         let mut cx = hs::ClientContext {
             common: &mut common_state,
             data: &mut data,
+            // `start_handshake` won't produce plaintext
+            sendable_plaintext: None,
         };
 
         let state = hs::start_handshake(name, extra_exts, config, &mut cx)?;
         Ok(Self::new(state, data, common_state))
     }
 
+    #[cfg(feature = "std")]
     pub(crate) fn is_early_data_accepted(&self) -> bool {
         self.data.early_data.is_accepted()
     }
 }
 
+/// Unbuffered version of `ClientConnection`
+///
+/// See the [`crate::unbuffered`] module docs for more details
+pub struct UnbufferedClientConnection {
+    inner: UnbufferedConnectionCommon<ClientConnectionData>,
+}
+
+impl UnbufferedClientConnection {
+    /// Make a new ClientConnection. `config` controls how we behave in the TLS protocol, `name` is
+    /// the name of the server we want to talk to.
+    pub fn new(config: Arc<ClientConfig>, name: ServerName<'static>) -> Result<Self, Error> {
+        Ok(Self {
+            inner: ConnectionCore::for_client(config, name, Vec::new(), Protocol::Tcp)?.into(),
+        })
+    }
+}
+
+impl Deref for UnbufferedClientConnection {
+    type Target = UnbufferedConnectionCommon<ClientConnectionData>;
+
+    fn deref(&self) -> &Self::Target {
+        &self.inner
+    }
+}
+
+impl DerefMut for UnbufferedClientConnection {
+    fn deref_mut(&mut self) -> &mut Self::Target {
+        &mut self.inner
+    }
+}
+
+impl TransmitTlsData<'_, ClientConnectionData> {
+    /// returns an adapter that allows encrypting early (RTT-0) data before transmitting the
+    /// already encoded TLS data
+    ///
+    /// IF allowed by the protocol
+    pub fn may_encrypt_early_data(&mut self) -> Option<MayEncryptEarlyData<'_>> {
+        if self
+            .conn
+            .core
+            .data
+            .early_data
+            .is_enabled()
+        {
+            Some(MayEncryptEarlyData { conn: self.conn })
+        } else {
+            None
+        }
+    }
+}
+
+/// Allows encrypting early (RTT-0) data
+pub struct MayEncryptEarlyData<'c> {
+    conn: &'c mut UnbufferedConnectionCommon<ClientConnectionData>,
+}
+
+impl MayEncryptEarlyData<'_> {
+    /// Encrypts `application_data` into the `outgoing_tls` buffer
+    ///
+    /// returns the number of bytes that were written into `outgoing_tls`, or an error if
+    /// the provided buffer was too small. In the error case, `outgoing_tls` is not modified
+    pub fn encrypt(
+        &mut self,
+        early_data: &[u8],
+        outgoing_tls: &mut [u8],
+    ) -> Result<usize, EarlyDataError> {
+        let allowed = match self
+            .conn
+            .core
+            .data
+            .early_data
+            .check_write_opt(early_data.len())
+        {
+            Some(allowed) => allowed,
+            None => return Err(EarlyDataError::ExceededAllowedEarlyData),
+        };
+
+        self.conn
+            .core
+            .common_state
+            .write_plaintext(early_data[..allowed].into(), outgoing_tls)
+            .map_err(|e| e.into())
+    }
+}
+
+/// Errors that may arise when encrypting early (RTT-0) data
+#[derive(Debug)]
+pub enum EarlyDataError {
+    /// Cannot encrypt more early data due to imposed limits
+    ExceededAllowedEarlyData,
+    /// Encryption error
+    Encrypt(EncryptError),
+}
+
+impl From<EncryptError> for EarlyDataError {
+    fn from(v: EncryptError) -> Self {
+        Self::Encrypt(v)
+    }
+}
+
+impl fmt::Display for EarlyDataError {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        match self {
+            Self::ExceededAllowedEarlyData => f.write_str("cannot send any more early data"),
+            Self::Encrypt(e) => fmt::Display::fmt(e, f),
+        }
+    }
+}
+
+#[cfg(feature = "std")]
+impl std::error::Error for EarlyDataError {}
+
 /// State associated with a client connection.
 #[derive(Debug)]
 pub struct ClientConnectionData {
     pub(super) early_data: EarlyData,
     pub(super) resumption_ciphersuite: Option<SupportedCipherSuite>,
+    pub(super) ech_status: EchStatus,
+    pub(super) fips: bool,
 }
 
 impl ClientConnectionData {
@@ -672,6 +954,8 @@ impl ClientConnectionData {
         Self {
             early_data: EarlyData::new(),
             resumption_ciphersuite: None,
+            ech_status: EchStatus::NotOffered,
+            fips: false,
         }
     }
 }
