@@ -34,14 +34,14 @@ impl client::ClientSessionStore for NoClientSessionStorage {
     }
 }
 
-#[cfg(feature = "std")]
+#[cfg(any(feature = "std", feature = "hashbrown"))]
 mod cache {
     use alloc::collections::VecDeque;
     use core::fmt;
-    use std::sync::Mutex;
 
     use pki_types::ServerName;
 
+    use crate::lock::Mutex;
     use crate::msgs::persist;
     use crate::{limited_cache, NamedGroup};
 
@@ -80,6 +80,7 @@ mod cache {
     impl ClientSessionMemoryCache {
         /// Make a new ClientSessionMemoryCache.  `size` is the
         /// maximum number of stored sessions.
+        #[cfg(feature = "std")]
         pub fn new(size: usize) -> Self {
             let max_servers = size.saturating_add(MAX_TLS13_TICKETS_PER_SERVER - 1)
                 / MAX_TLS13_TICKETS_PER_SERVER;
@@ -87,6 +88,17 @@ mod cache {
                 servers: Mutex::new(limited_cache::LimitedCache::new(max_servers)),
             }
         }
+
+        /// Make a new ClientSessionMemoryCache.  `size` is the
+        /// maximum number of stored sessions.
+        #[cfg(not(feature = "std"))]
+        pub fn new<M: crate::lock::MakeMutex>(size: usize) -> Self {
+            let max_servers = size.saturating_add(MAX_TLS13_TICKETS_PER_SERVER - 1)
+                / MAX_TLS13_TICKETS_PER_SERVER;
+            Self {
+                servers: Mutex::new::<M>(limited_cache::LimitedCache::new(max_servers)),
+            }
+        }
     }
 
     impl super::client::ClientSessionStore for ClientSessionMemoryCache {
@@ -180,7 +192,7 @@ mod cache {
     }
 }
 
-#[cfg(feature = "std")]
+#[cfg(any(feature = "std", feature = "hashbrown"))]
 pub use cache::ClientSessionMemoryCache;
 
 #[derive(Debug)]
@@ -231,6 +243,7 @@ impl client::ResolvesClientCert for AlwaysResolvesClientCert {
 
 test_for_each_provider! {
     use std::prelude::v1::*;
+    use alloc::sync::Arc;
     use super::NoClientSessionStorage;
     use crate::client::ClientSessionStore;
     use crate::msgs::enums::NamedGroup;
@@ -239,6 +252,7 @@ test_for_each_provider! {
     use crate::msgs::handshake::SessionId;
     use crate::msgs::persist::Tls13ClientSessionValue;
     use crate::suites::SupportedCipherSuite;
+    use crate::msgs::base::PayloadU16;
     use provider::cipher_suite;
 
     use pki_types::{ServerName, UnixTime};
@@ -266,7 +280,7 @@ test_for_each_provider! {
                 Tls12ClientSessionValue::new(
                     tls12_suite,
                     SessionId::empty(),
-                    Vec::new(),
+                    Arc::new(PayloadU16::empty()),
                     &[],
                     CertificateChain::default(),
                     now,
@@ -288,7 +302,7 @@ test_for_each_provider! {
             name.clone(),
             Tls13ClientSessionValue::new(
                 tls13_suite,
-                Vec::new(),
+                Arc::new(PayloadU16::empty()),
                 &[],
                 CertificateChain::default(),
                 now,
