@@ -126,7 +126,7 @@ pub(super) fn start_handshake(
                 // If we have a ticket, we use the sessionid as a signal that
                 // we're  doing an abbreviated handshake.  See section 3.4 in
                 // RFC5077.
-                if !inner.ticket().is_empty() {
+                if !inner.ticket().0.is_empty() {
                     inner.session_id = SessionId::random(config.provider.secure_random)?;
                 }
                 Some(inner.session_id)
@@ -242,16 +242,22 @@ fn emit_client_hello_for_retry(
     // should be unreachable thanks to config builder
     assert!(!supported_versions.is_empty());
 
+    // offer groups which are usable for any offered version
+    let offered_groups = config
+        .provider
+        .kx_groups
+        .iter()
+        .filter(|skxg| {
+            supported_versions
+                .iter()
+                .any(|v| skxg.usable_for_version(*v))
+        })
+        .map(|skxg| skxg.name())
+        .collect();
+
     let mut exts = vec![
         ClientExtension::SupportedVersions(supported_versions),
-        ClientExtension::NamedGroups(
-            config
-                .provider
-                .kx_groups
-                .iter()
-                .map(|skxg| skxg.name())
-                .collect(),
-        ),
+        ClientExtension::NamedGroups(offered_groups),
         ClientExtension::SignatureAlgorithms(
             config
                 .verifier
@@ -1028,7 +1034,7 @@ impl ExpectServerHelloOrHelloRetryRequest {
 
         let key_share = match req_group {
             Some(group) if group != offered_key_share.group() => {
-                let skxg = match config.find_kx_group(group) {
+                let skxg = match config.find_kx_group(group, ProtocolVersion::TLSv1_3) {
                     Some(skxg) => skxg,
                     None => {
                         return Err(cx.common.send_fatal_alert(
