@@ -15,19 +15,22 @@ use crate::rand::GetRandomFailed;
 use crate::sign::SigningKey;
 use crate::suites::SupportedCipherSuite;
 use crate::webpki::WebPkiSupportedAlgorithms;
-use crate::Error;
+use crate::{Error, OtherError};
 
+/// Hybrid public key encryption (HPKE).
+pub mod hpke;
 /// Using software keys for authentication.
 pub mod sign;
 
 #[path = "../ring/hash.rs"]
 pub(crate) mod hash;
+#[path = "../ring/hmac.rs"]
+pub(crate) mod hmac;
 #[path = "../ring/kx.rs"]
 pub(crate) mod kx;
 #[path = "../ring/quic.rs"]
 pub(crate) mod quic;
-#[cfg(feature = "std")]
-#[path = "../ring/ticketer.rs"]
+#[cfg(any(feature = "std", feature = "hashbrown"))]
 pub(crate) mod ticketer;
 #[cfg(feature = "tls12")]
 pub(crate) mod tls12;
@@ -114,7 +117,7 @@ pub static DEFAULT_CIPHER_SUITES: &[SupportedCipherSuite] = &[
     tls12::TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256,
 ];
 
-/// A list of all the cipher suites supported by the rustls *ring* provider.
+/// A list of all the cipher suites supported by the rustls aws-lc-rs provider.
 pub static ALL_CIPHER_SUITES: &[SupportedCipherSuite] = &[
     // TLS1.3 suites
     tls13::TLS13_AES_256_GCM_SHA384,
@@ -149,13 +152,15 @@ pub mod cipher_suite {
 }
 
 /// A `WebPkiSupportedAlgorithms` value that reflects webpki's capabilities when
-/// compiled against *ring*.
+/// compiled against aws-lc-rs.
 static SUPPORTED_SIG_ALGS: WebPkiSupportedAlgorithms = WebPkiSupportedAlgorithms {
     all: &[
         webpki_algs::ECDSA_P256_SHA256,
         webpki_algs::ECDSA_P256_SHA384,
         webpki_algs::ECDSA_P384_SHA256,
         webpki_algs::ECDSA_P384_SHA384,
+        webpki_algs::ECDSA_P521_SHA256,
+        webpki_algs::ECDSA_P521_SHA384,
         webpki_algs::ECDSA_P521_SHA512,
         webpki_algs::ED25519,
         webpki_algs::RSA_PSS_2048_8192_SHA256_LEGACY_KEY,
@@ -168,15 +173,12 @@ static SUPPORTED_SIG_ALGS: WebPkiSupportedAlgorithms = WebPkiSupportedAlgorithms
     ],
     mapping: &[
         // Note: for TLS1.2 the curve is not fixed by SignatureScheme. For TLS1.3 it is.
-        (
-            SignatureScheme::ECDSA_NISTP521_SHA512,
-            &[webpki_algs::ECDSA_P521_SHA512],
-        ),
         (
             SignatureScheme::ECDSA_NISTP384_SHA384,
             &[
                 webpki_algs::ECDSA_P384_SHA384,
                 webpki_algs::ECDSA_P256_SHA384,
+                webpki_algs::ECDSA_P521_SHA384,
             ],
         ),
         (
@@ -184,8 +186,13 @@ static SUPPORTED_SIG_ALGS: WebPkiSupportedAlgorithms = WebPkiSupportedAlgorithms
             &[
                 webpki_algs::ECDSA_P256_SHA256,
                 webpki_algs::ECDSA_P384_SHA256,
+                webpki_algs::ECDSA_P521_SHA256,
             ],
         ),
+        (
+            SignatureScheme::ECDSA_NISTP521_SHA512,
+            &[webpki_algs::ECDSA_P521_SHA512],
+        ),
         (SignatureScheme::ED25519, &[webpki_algs::ED25519]),
         (
             SignatureScheme::RSA_PSS_SHA512,
@@ -222,7 +229,7 @@ pub mod kx_group {
 }
 
 pub use kx::ALL_KX_GROUPS;
-#[cfg(feature = "std")]
+#[cfg(any(feature = "std", feature = "hashbrown"))]
 pub use ticketer::Ticketer;
 
 use super::SupportedKxGroup;
@@ -242,15 +249,22 @@ mod ring_shim {
     }
 }
 
-/// AEAD algorithm that is used by `mod ticketer`.
-#[cfg(feature = "std")]
-pub(super) static TICKETER_AEAD: &ring_like::aead::Algorithm = &ring_like::aead::AES_256_GCM;
-
 /// Are we in FIPS mode?
 pub(super) fn fips() -> bool {
     aws_lc_rs::try_fips_mode().is_ok()
 }
 
+pub(super) fn unspecified_err(_e: aws_lc_rs::error::Unspecified) -> Error {
+    #[cfg(feature = "std")]
+    {
+        Error::Other(OtherError(Arc::new(_e)))
+    }
+    #[cfg(not(feature = "std"))]
+    {
+        Error::Other(OtherError())
+    }
+}
+
 #[cfg(test)]
 mod tests {
     #[cfg(feature = "fips")]
