@@ -1,14 +1,19 @@
+use alloc::boxed::Box;
+use alloc::vec::Vec;
+use core::mem;
+
+use crate::crypto::hash;
 use crate::msgs::codec::Codec;
+use crate::msgs::enums::HashAlgorithm;
 use crate::msgs::handshake::HandshakeMessagePayload;
 use crate::msgs::message::{Message, MessagePayload};
-use ring::digest;
-use std::mem;
 
 /// Early stage buffering of handshake payloads.
 ///
 /// Before we know the hash algorithm to use to verify the handshake, we just buffer the messages.
 /// During the handshake, we may restart the transcript due to a HelloRetryRequest, reverting
 /// from the `HandshakeHash` to a `HandshakeHashBuffer` again.
+#[derive(Clone)]
 pub(crate) struct HandshakeHashBuffer {
     buffer: Vec<u8>,
     client_auth_enabled: bool,
@@ -29,36 +34,37 @@ impl HandshakeHashBuffer {
     }
 
     /// Hash/buffer a handshake message.
-    pub(crate) fn add_message(&mut self, m: &Message) {
-        if let MessagePayload::Handshake(hs) = &m.payload {
-            self.buffer
-                .extend_from_slice(&hs.get_encoding());
-        }
+    pub(crate) fn add_message(&mut self, m: &Message<'_>) {
+        match &m.payload {
+            MessagePayload::Handshake { encoded, .. } => self.add_raw(encoded.bytes()),
+            MessagePayload::HandshakeFlight(payload) => self.add_raw(payload.bytes()),
+            _ => {}
+        };
     }
 
     /// Hash or buffer a byte slice.
-    #[cfg(test)]
-    fn update_raw(&mut self, buf: &[u8]) {
+    fn add_raw(&mut self, buf: &[u8]) {
         self.buffer.extend_from_slice(buf);
     }
 
     /// Get the hash value if we were to hash `extra` too.
-    pub(crate) fn get_hash_given(
+    pub(crate) fn hash_given(
         &self,
-        hash: &'static digest::Algorithm,
+        provider: &'static dyn hash::Hash,
         extra: &[u8],
-    ) -> digest::Digest {
-        let mut ctx = digest::Context::new(hash);
+    ) -> hash::Output {
+        let mut ctx = provider.start();
         ctx.update(&self.buffer);
         ctx.update(extra);
         ctx.finish()
     }
 
     /// We now know what hash function the verify_data will use.
-    pub(crate) fn start_hash(self, alg: &'static digest::Algorithm) -> HandshakeHash {
-        let mut ctx = digest::Context::new(alg);
+    pub(crate) fn start_hash(self, provider: &'static dyn hash::Hash) -> HandshakeHash {
+        let mut ctx = provider.start();
         ctx.update(&self.buffer);
         HandshakeHash {
+            provider,
             ctx,
             client_auth: match self.client_auth_enabled {
                 true => Some(self.buffer),
@@ -76,8 +82,8 @@ impl HandshakeHashBuffer {
 /// For client auth, we also need to buffer all the messages.
 /// This is disabled in cases where client auth is not possible.
 pub(crate) struct HandshakeHash {
-    /// None before we know what hash function we're using
-    ctx: digest::Context,
+    provider: &'static dyn hash::Hash,
+    ctx: Box<dyn hash::Context>,
 
     /// buffer for client-auth.
     client_auth: Option<Vec<u8>>,
@@ -91,16 +97,21 @@ impl HandshakeHash {
     }
 
     /// Hash/buffer a handshake message.
-    pub(crate) fn add_message(&mut self, m: &Message) -> &mut Self {
-        if let MessagePayload::Handshake(hs) = &m.payload {
-            let buf = hs.get_encoding();
-            self.update_raw(&buf);
+    pub(crate) fn add_message(&mut self, m: &Message<'_>) -> &mut Self {
+        match &m.payload {
+            MessagePayload::Handshake { encoded, .. } => self.add_raw(encoded.bytes()),
+            MessagePayload::HandshakeFlight(payload) => self.add_raw(payload.bytes()),
+            _ => self,
         }
-        self
+    }
+
+    /// Hash/buffer an encoded handshake message.
+    pub(crate) fn add(&mut self, bytes: &[u8]) {
+        self.add_raw(bytes);
     }
 
     /// Hash or buffer a byte slice.
-    fn update_raw(&mut self, buf: &[u8]) -> &mut Self {
+    fn add_raw(&mut self, buf: &[u8]) -> &mut Self {
         self.ctx.update(buf);
 
         if let Some(buffer) = &mut self.client_auth {
@@ -112,8 +123,8 @@ impl HandshakeHash {
 
     /// Get the hash value if we were to hash `extra` too,
     /// using hash function `hash`.
-    pub(crate) fn get_hash_given(&self, extra: &[u8]) -> digest::Digest {
-        let mut ctx = self.ctx.clone();
+    pub(crate) fn hash_given(&self, extra: &[u8]) -> hash::Output {
+        let mut ctx = self.ctx.fork();
         ctx.update(extra);
         ctx.finish()
     }
@@ -135,17 +146,17 @@ impl HandshakeHash {
     pub(crate) fn rollup_for_hrr(&mut self) {
         let ctx = &mut self.ctx;
 
-        let old_ctx = mem::replace(ctx, digest::Context::new(ctx.algorithm()));
+        let old_ctx = mem::replace(ctx, self.provider.start());
         let old_hash = old_ctx.finish();
         let old_handshake_hash_msg =
             HandshakeMessagePayload::build_handshake_hash(old_hash.as_ref());
 
-        self.update_raw(&old_handshake_hash_msg.get_encoding());
+        self.add_raw(&old_handshake_hash_msg.get_encoding());
     }
 
     /// Get the current hash value.
-    pub(crate) fn get_current_hash(&self) -> digest::Digest {
-        self.ctx.clone().finish()
+    pub(crate) fn current_hash(&self) -> hash::Output {
+        self.ctx.fork_finish()
     }
 
     /// Takes this object's buffer containing all handshake messages
@@ -156,26 +167,39 @@ impl HandshakeHash {
         self.client_auth.take()
     }
 
-    /// The digest algorithm
-    pub(crate) fn algorithm(&self) -> &'static digest::Algorithm {
-        self.ctx.algorithm()
+    /// The hashing algorithm
+    pub(crate) fn algorithm(&self) -> HashAlgorithm {
+        self.provider.algorithm()
     }
 }
 
-#[cfg(test)]
-mod test {
-    use super::HandshakeHashBuffer;
-    use ring::digest;
+impl Clone for HandshakeHash {
+    fn clone(&self) -> Self {
+        Self {
+            provider: self.provider,
+            ctx: self.ctx.fork(),
+            client_auth: self.client_auth.clone(),
+        }
+    }
+}
+
+test_for_each_provider! {
+    use super::*;
+    use crate::msgs::handshake::{HandshakeMessagePayload, HandshakePayload};
+    use crate::crypto::hash::Hash;
+    use crate::msgs::base::Payload;
+    use crate::enums::{ProtocolVersion, HandshakeType};
+    use provider::hash::SHA256;
 
     #[test]
     fn hashes_correctly() {
         let mut hhb = HandshakeHashBuffer::new();
-        hhb.update_raw(b"hello");
+        hhb.add_raw(b"hello");
         assert_eq!(hhb.buffer.len(), 5);
-        let mut hh = hhb.start_hash(&digest::SHA256);
+        let mut hh = hhb.start_hash(&SHA256);
         assert!(hh.client_auth.is_none());
-        hh.update_raw(b"world");
-        let h = hh.get_current_hash();
+        hh.add_raw(b"world");
+        let h = hh.current_hash();
         let h = h.as_ref();
         assert_eq!(h[0], 0x93);
         assert_eq!(h[1], 0x6a);
@@ -183,28 +207,66 @@ mod test {
         assert_eq!(h[3], 0x5c);
     }
 
+    #[test]
+    fn hashes_message_types() {
+        // handshake protocol encoding of 0x0e 00 00 00
+        let server_hello_done_message = Message {
+            version: ProtocolVersion::TLSv1_2,
+            payload: MessagePayload::handshake(HandshakeMessagePayload {
+                        typ: HandshakeType::ServerHelloDone,
+                        payload: HandshakePayload::ServerHelloDone,
+                        }) };
+
+        let app_data_ignored = Message {
+            version: ProtocolVersion::TLSv1_3,
+            payload: MessagePayload::ApplicationData(Payload::Borrowed(b"hello")),
+        };
+
+        let end_of_early_data_flight = Message {
+            version: ProtocolVersion::TLSv1_3,
+            payload: MessagePayload::HandshakeFlight(Payload::Borrowed(b"\x05\x00\x00\x00"))
+        };
+
+        // buffered mode
+        let mut hhb = HandshakeHashBuffer::new();
+        hhb.add_message(&server_hello_done_message);
+        hhb.add_message(&app_data_ignored);
+        hhb.add_message(&end_of_early_data_flight);
+        assert_eq!(hhb.start_hash(&SHA256).current_hash().as_ref(),
+                   SHA256.hash(b"\x0e\x00\x00\x00\x05\x00\x00\x00").as_ref());
+
+        // non-buffered mode
+        let mut hh = HandshakeHashBuffer::new().start_hash(&SHA256);
+        hh.add_message(&server_hello_done_message);
+        hh.add_message(&app_data_ignored);
+        hh.add_message(&end_of_early_data_flight);
+        assert_eq!(hh.current_hash().as_ref(),
+                   SHA256.hash(b"\x0e\x00\x00\x00\x05\x00\x00\x00").as_ref());
+
+    }
+
     #[cfg(feature = "tls12")]
     #[test]
     fn buffers_correctly() {
         let mut hhb = HandshakeHashBuffer::new();
         hhb.set_client_auth_enabled();
-        hhb.update_raw(b"hello");
+        hhb.add_raw(b"hello");
         assert_eq!(hhb.buffer.len(), 5);
-        let mut hh = hhb.start_hash(&digest::SHA256);
+        let mut hh = hhb.start_hash(&SHA256);
         assert_eq!(
             hh.client_auth
                 .as_ref()
                 .map(|buf| buf.len()),
             Some(5)
         );
-        hh.update_raw(b"world");
+        hh.add_raw(b"world");
         assert_eq!(
             hh.client_auth
                 .as_ref()
                 .map(|buf| buf.len()),
             Some(10)
         );
-        let h = hh.get_current_hash();
+        let h = hh.current_hash();
         let h = h.as_ref();
         assert_eq!(h[0], 0x93);
         assert_eq!(h[1], 0x6a);
@@ -218,9 +280,9 @@ mod test {
     fn abandon() {
         let mut hhb = HandshakeHashBuffer::new();
         hhb.set_client_auth_enabled();
-        hhb.update_raw(b"hello");
+        hhb.add_raw(b"hello");
         assert_eq!(hhb.buffer.len(), 5);
-        let mut hh = hhb.start_hash(&digest::SHA256);
+        let mut hh = hhb.start_hash(&SHA256);
         assert_eq!(
             hh.client_auth
                 .as_ref()
@@ -229,13 +291,46 @@ mod test {
         );
         hh.abandon_client_auth();
         assert_eq!(hh.client_auth, None);
-        hh.update_raw(b"world");
+        hh.add_raw(b"world");
         assert_eq!(hh.client_auth, None);
-        let h = hh.get_current_hash();
+        let h = hh.current_hash();
         let h = h.as_ref();
         assert_eq!(h[0], 0x93);
         assert_eq!(h[1], 0x6a);
         assert_eq!(h[2], 0x18);
         assert_eq!(h[3], 0x5c);
     }
+
+    #[test]
+    fn clones_correctly() {
+        let mut hhb = HandshakeHashBuffer::new();
+        hhb.set_client_auth_enabled();
+        hhb.add_raw(b"hello");
+        assert_eq!(hhb.buffer.len(), 5);
+
+        // Cloning the HHB should result in the same buffer and client auth state.
+        let mut hhb_prime = hhb.clone();
+        assert_eq!(hhb_prime.buffer, hhb.buffer);
+        assert!(hhb_prime.client_auth_enabled);
+
+        // Updating the HHB clone shouldn't affect the original.
+        hhb_prime.add_raw(b"world");
+        assert_eq!(hhb_prime.buffer.len(), 10);
+        assert_ne!(hhb.buffer, hhb_prime.buffer);
+
+        let hh = hhb.start_hash(&SHA256);
+        let hh_hash = hh.current_hash();
+        let hh_hash = hh_hash.as_ref();
+
+        // Cloning the HH should result in the same current hash.
+        let mut hh_prime = hh.clone();
+        let hh_prime_hash = hh_prime.current_hash();
+        let hh_prime_hash = hh_prime_hash.as_ref();
+        assert_eq!(hh_hash, hh_prime_hash);
+
+        // Updating the HH clone shouldn't affect the original.
+        hh_prime.add_raw(b"goodbye");
+        assert_eq!(hh.current_hash().as_ref(), hh_hash);
+        assert_ne!(hh_prime.current_hash().as_ref(), hh_hash);
+    }
 }
