@@ -2,7 +2,6 @@ use alloc::sync::Arc;
 use alloc::vec::Vec;
 use core::fmt::Debug;
 
-use crate::msgs::handshake::CertificateChain;
 use crate::server::ClientHello;
 use crate::{server, sign};
 
@@ -172,23 +171,16 @@ impl server::ProducesTickets for NeverProducesTickets {
 pub(super) struct AlwaysResolvesChain(Arc<sign::CertifiedKey>);
 
 impl AlwaysResolvesChain {
-    /// Creates an `AlwaysResolvesChain`, using the supplied key and certificate chain.
-    pub(super) fn new(
-        private_key: Arc<dyn sign::SigningKey>,
-        chain: CertificateChain<'static>,
-    ) -> Self {
-        Self(Arc::new(sign::CertifiedKey::new(chain.0, private_key)))
+    /// Creates an `AlwaysResolvesChain`, using the supplied `CertifiedKey`.
+    pub(super) fn new(certified_key: sign::CertifiedKey) -> Self {
+        Self(Arc::new(certified_key))
     }
 
-    /// Creates an `AlwaysResolvesChain`, using the supplied key, certificate chain and OCSP response.
+    /// Creates an `AlwaysResolvesChain`, using the supplied `CertifiedKey` and OCSP response.
     ///
     /// If non-empty, the given OCSP response is attached.
-    pub(super) fn new_with_extras(
-        private_key: Arc<dyn sign::SigningKey>,
-        chain: CertificateChain<'static>,
-        ocsp: Vec<u8>,
-    ) -> Self {
-        let mut r = Self::new(private_key, chain);
+    pub(super) fn new_with_extras(certified_key: sign::CertifiedKey, ocsp: Vec<u8>) -> Self {
+        let mut r = Self::new(certified_key);
 
         {
             let cert = Arc::make_mut(&mut r.0);
@@ -202,7 +194,7 @@ impl AlwaysResolvesChain {
 }
 
 impl server::ResolvesServerCert for AlwaysResolvesChain {
-    fn resolve(&self, _client_hello: ClientHello) -> Option<Arc<sign::CertifiedKey>> {
+    fn resolve(&self, _client_hello: ClientHello<'_>) -> Option<Arc<sign::CertifiedKey>> {
         Some(Arc::clone(&self.0))
     }
 }
@@ -271,7 +263,7 @@ mod sni_resolver {
     }
 
     impl server::ResolvesServerCert for ResolvesServerCertUsingSni {
-        fn resolve(&self, client_hello: ClientHello) -> Option<Arc<sign::CertifiedKey>> {
+        fn resolve(&self, client_hello: ClientHello<'_>) -> Option<Arc<sign::CertifiedKey>> {
             if let Some(name) = client_hello.server_name() {
                 self.by_name.get(name).cloned()
             } else {
