@@ -430,7 +430,9 @@ impl ExpectClientHello {
                 .provider
                 .kx_groups
                 .iter()
-                .find(|skxg| skxg.name() == *offered_group);
+                .find(|skxg| {
+                    skxg.usable_for_version(selected_version) && skxg.name() == *offered_group
+                });
 
             match offered_group.key_exchange_algorithm() {
                 KeyExchangeAlgorithm::DHE => {
